from .modelling import ArithmeticMixin

class Pipeline(ArithmeticMixin):
    def __init__(self, model, featureslice):
        self.model = model
        self.featureslice = featureslice
        self.__name__ = self.model.__name__ + '-' + self.featureslice.featureset.__name__

    def fit(self, **kwargs):
        X = self.featureslice().values
        y = self.featureslice.target.values
        self.model.fit(X, y, **kwargs)

    def predict(self, df, **kwargs):
        return self.model.predict(self.featureslice(df).values, **kwargs)


import os
import sys
sys.path.insert(0, '.')


from .cli import check_file_system
check_file_system()
from .feature.decorators import preview, register, deregister, dropper, selector, helper
from .feature import stl
from .feature.storage import feature_list as features
from .feature import helper_list as helpers
from .feature.storage import FeatureSet
from .validation.experiment import experiment_list as experiments
from .validation import Validator
from .modelling import model
from .storage.dataframe import link
from .storage import cache
from .storage.dataframe import DataFrame as KTDF
from .storage.caching import save, load, ls, remove, rm
from .validation.leaderboard import leaderboard
lb = leaderboard
from . import zoo

# import mprop
#
# @property
# def __version__(kts):
#     ans = os.popen(f'{sys.executable} -m pip show kts').read().split()
#     return [ans[i + 1] for i in range(len(ans)) if 'version' in ans[i].lower()][0]
#
# mprop.init()


from ..modelling import ArithmeticMixin
from ..storage import cache
import re
from collections import MutableSequence
from ..utils import hash_str
import pandas as pd
import texttable as tt


class Experiment(ArithmeticMixin):
    def __init__(self, pipeline, oofs, score, std, description, splitter, metric):
        self.pipeline = pipeline
        self.model = self.pipeline.models[0].model.model  # TODO: test out
        self.model_name = self.model.__class__.__name__
        self.parameters = self.model.get_params()
        self.models = [model.model.model for model in self.pipeline.models]
        self.featureset = self.pipeline.models[0].model.featureslice.featureset
        self.oofs = oofs
        self.score = score
        self.std = std
        self.identifier = hash_str(f'{round(score, 4)}-{pipeline.__name__}')[:6].upper()
        self.oofs.columns = [self.identifier]
        self.__doc__ = description if description is not None else 'no description'
        self.__name__ = f"{self.identifier}-{round(score, 4)}-{pipeline.__name__}"
        self.splitter = splitter
        self.metric = metric

    def __str__(self):
        string = f"({round(self.score, 5)}, std:{round(self.std, 3)}: \n\tModel: {self.pipeline.__name__})"
        return string

    def predict(self, df):
        return self.pipeline.predict(df)

    def __repr__(self):
        fields = {
            'Score': f"{round(self.score, 7)}, std: {round(self.std, 7)} ({self.metric.__name__})",
            'Identifier': self.identifier,
            'Description': self.__doc__,
            # 'Model': self.model_name + f'\tx{len(self.models)}',
            # 'Model parameters': self.parameters,
            'Model': f"{self.model.__name__}\t x{len(self.models)}",
            '|- source ': self.model.source,              # be careful with refactoring: if you remove this space,
            'FeatureSet': self.featureset.__name__,       #
            '|- description': self.featureset.__doc__,    #
            '|- source': self.featureset.source,          # both "source" rows will be considered identical
            'Splitter': self.splitter,
        }

        table = tt.Texttable(max_width=80)
        for field in fields:
            table.add_row([field, fields[field]])
        return table.draw()

    def as_dict(self):
        fields = {
            'Score': self.score,
            'std': self.std,
            'ID': self.identifier,
            'Model': self.model.__name__,
            'FS': self.featureset.__name__,
            'Description': self.__doc__,
            'FS description': self.featureset.__doc__,
            'Model source': self.model.source,
            'FS source': self.featureset.source,
            'Splitter': repr(self.splitter)
        }
        for field in fields:
            fields[field] = [fields[field]]
        return fields

    def as_df(self):
        return pd.DataFrame(self.as_dict()).set_index('ID')


class ExperimentList(MutableSequence):
    def __init__(self):
        self.experiments = []
        self.name_to_idx = dict()

    def recalc(self):
        self.experiments = []
        self.name_to_idx = dict()
        names = [obj for obj in cache.cached_objs() if obj.endswith('_exp')]
        for idx, name in enumerate(names):
            # print(idx, name)
            experiment = cache.load_obj(name)
            self.experiments.append(experiment)
            self.name_to_idx[experiment.__name__] = idx
        self.experiments.sort(key=lambda e: e.score, reverse=True)

    def __getitem__(self, item):
        """
        Implements calling to experiments by score, name and index
        :param item: str(name) or float(score), slice(score range or index range)
        :return: experiment or list of experiments
        """
        raise DeprecationWarning('Use kts.lb, kts.experiments is deprecated')
        self.recalc()
        if isinstance(item, str):
            if bool(re.match('[0-9A-F]{6}', item)):
                ans = [experiment
                       for experiment in self.experiments
                       if 'identifier' in dir(experiment)
                       and experiment.identifier == item]
            else:
                ans = [experiment for experiment in self.experiments if experiment.__name__.count(item) > 0]
        elif isinstance(item, float):
            mul = 10 ** len(str(item).split('.')[1])
            ans = [experiment for experiment in self.experiments if int(experiment.score * mul) == int(item * mul)]
        elif isinstance(item, int):
            return self.experiments[item]
        elif isinstance(item, slice):
            if type(item.start) == float:
                ans = [experiment for experiment in self.experiments if
                       item.start <= experiment.score and experiment.score < item.stop]
            else:
                return self.experiments[item]
        else:
            raise TypeError("Item must be of str, number or slice type")

        if len(ans) > 1:
            return ans
        elif len(ans) == 1:
            return ans[0]
        else:
            return

    def __repr__(self):
        raise DeprecationWarning('Use kts.lb, kts.experiments is deprecated')
        self.recalc()
        string = "Experiments: [\n" + '\n'.join([experiment.__str__() for experiment in self.experiments]) + '\n]'
        return string

    def __delitem__(self, key):
        raise AttributeError('This object is read-only')

    def __setitem__(self, key, value):
        raise AttributeError('This object is read-only')

    def insert(self, key, value):
        raise AttributeError('This object is read-only')

    def register(self, experiment):
        cache.cache_obj(experiment, experiment.__name__ + '_exp')

    def __len__(self):
        raise DeprecationWarning('Use kts.lb, kts.experiments is deprecated')
        self.recalc()
        return len(self.experiments)


experiment_list = ExperimentList()



import numpy as np
from hashlib import sha256
import json


class NamingMixin:
    @property
    def __name__(self):
        self.params = {key: self.get_params()[key] for key in self.tracked_params if
                       key in self.get_params()}
        return f"{self.short_name}_{sha256((json.dumps(self.params, sort_keys=True)).encode()).hexdigest()[-3:] if self.params else 'default'}"


class ArithmeticMixin:
    def __mul__(self, x):
        return WeightedModel(self, x)

    def __rmul__(self, x):
        return self * x

    def __truediv__(self, x):
        assert x != 0
        return WeightedModel(self, 1. / x)

    def __add__(self, other):
        # assert type(self).__base__ == type(other).__base__, \
        # f"Can't add {type(other).__name__} to {type(self).__name__}."
        if other == 0:
            return self
        return Ensemble([self, other])

    def __radd__(self, other):
        return self + other


class SourceMixin:
    @property
    def source(self):
        args = []
        for key, value in self.get_params().items():
            args.append(f"{key}={repr(value)}")
        res = ', '.join(args)
        return f"{self.__class__.__name__}({res})"


class Model(ArithmeticMixin, NamingMixin, SourceMixin):
    tracked_params = []
    short_name = 'model'

    def __repr__(self):
        return self.__name__

class WeightedModel(ArithmeticMixin): # MulNode
    """
    Multiplies predictions by a certain coefficient.
    """
    def __init__(self, model, coeff):
        self.model = model
        self.coeff = coeff
        if type(self.model) == type(self):
            self.coeff *= self.model.coeff
            self.model = self.model.model
        self.__name__ = f"{round(self.coeff, 2)} * "
        if '+' in self.model.__name__:
            self.__name__ += f"({self.model.__name__})"
        else:
            self.__name__ += f"{self.model.__name__}"

    def predict(self, X):
        """
        Standard prediction method.
        :param X: data matrix
        :return:
        """
        return self.coeff * self.model.predict(X)


class Ensemble(ArithmeticMixin): # AddNode
    """
    Sums up all predictions of all models.
    """
    def __init__(self, models):
        self.models = []
        ensembles = [model for model in models if type(model) == type(self)]
        single_models = [model for model in models if type(model) != type(self)]
        _models = []
        for ensemble in ensembles:
            _models += ensemble.models
        _models += single_models

        def __get_model_coeff(model):
            """
            :param model:
            :return: if the model is weighted, then returns its coefficient, else 1
            """
            if isinstance(model, WeightedModel):
                return model.model, model.coeff
            return model, 1

        self.models = list(set([__get_model_coeff(model)[0] for model in _models]))
        _coeffs = np.zeros(len(self.models))
        for i in range(len(self.models)):
            for model in _models:
                if __get_model_coeff(self.models[i])[0] == __get_model_coeff(model)[0]:
                    _coeffs[i] += __get_model_coeff(model)[1]

        self.models = [WeightedModel(model, coeff) for model, coeff in zip(self.models, _coeffs)]
        self.norm_coeff = _coeffs.sum()
        _coeffs /= self.norm_coeff
        self.__name__ = ' + '.join([(model / self.norm_coeff).__name__ for model in self.models])

    def predict(self, X):
        """
        Standard prediction method.
        :param X: data matrix
        :return:
        """
        res = 0
        for model in self.models:
            res += model.predict(X)
        return res / self.norm_coeff

    def __mul__(self, x):
        return Ensemble([model * x for model in self.models])

    def __truediv__(self, x):
        assert x != 0
        return Ensemble([model / x for model in self.models])


class CustomModel(ArithmeticMixin):
    def __init__(self, model, tracked_params=[], short_name=None, task='bc'):
        self.estimator = model
        self.short_name = (short_name if short_name is not None else self.get_short_name(model))
        self.tracked_params = tracked_params
        self.fit = self.estimator.fit
        self.task = task

    @property
    def __name__(self):
        self.params = {key: self.estimator.get_params()[key] for key in self.tracked_params if
                       key in self.estimator.get_params()}
        return f"{self.short_name}_{sha256((json.dumps(self.params, sort_keys=True)).encode()).hexdigest()[-2:] if self.params else 'default'}"

    @staticmethod
    def get_short_name(instance):
        return instance.__class__.__name__.lower().replace('classifier', '_clf').replace('regressor', '_rg')

    def __dir__(self):
        return set(dir(super())) | set(dir(self.estimator))

    def __getattr__(self, key):
        return getattr(self.estimator, key)

    def __setattr__(self, key, value):
        if 'estimator' in self.__dict__ and key in dir(self.estimator):
            setattr(self.estimator, key, value)
        else:
            super().__setattr__(key, value)

    def predict(self, X, **kw):
        if self.task == 'bc':
            return self.estimator.predict_proba(X, **kw)[:, 1]
        elif self.task == 'c':
            return self.estimator.predict_proba(X, **kw)
        elif self.task == 'r':
            return self.estimator.predict(X, **kw)


def model(model, tracked_params=[], short_name=None, task='bc'):
    """
    Creates a custom but trackable model.
    :param model: model object
    :param tracked_params: parameters to be tracked
    :param short_name: short name for your model, like 'rf' for RandomForestClassifier
    :param task: 'bc' for binary classification,
                 'c'  for classification,
                 'r'  for regression
    :return:
    """
    return CustomModel(model, tracked_params, short_name, task)


from .. import config
from ..storage import source_utils
from ..storage import cache_utils
from ..storage import caching
from ..storage import dataframe
import glob
import os
import numpy as np
import inspect


class FeatureConstructor:
    def __init__(self, function, cache_default=True, stl=False):
        self.function = function
        self.cache_default = cache_default
        self.__name__ = function.__name__
        self.source = source_utils.get_source(function)
        self.stl = stl

    # needs refactoring because of direct storing source
    def __call__(self, df, cache=None, **kwargs):
        if not self.stl:
            self = caching.cache.load_obj(self.__name__ + '_fc')
        # print("before constr", df.train)
        ktdf = dataframe.DataFrame(df)  # TODO: uncomment this line after tests with @preview
        # print("after constr:", ktdf.train)
        if type(cache) == type(None):
            cache = self.cache_default
        if not cache or config.preview_call:  # dirty hack to avoid  caching when @test function uses @registered function inside
            return self.function(ktdf, **kwargs)

        name = f"{self.function.__name__}__{cache_utils.get_hash_df(ktdf)[:4]}__{ktdf.slice_id[-4:]}"
        name_metadata = name + "_meta"
        if caching.cache.is_cached_df(name):
            if caching.cache.is_cached_obj(name_metadata):
                cached_encoders = caching.cache.load_obj(name_metadata)
                for key, value in cached_encoders.items():
                    ktdf.encoders[key] = value
            return dataframe.DataFrame(caching.cache.load_df(name), ktdf.train, ktdf.encoders, ktdf.slice_id)
        else:
            result = self.function(ktdf)
            caching.cache.cache_df(result, name)
            if ktdf.encoders:
                caching.cache.cache_obj(ktdf.encoders, name_metadata)
            return dataframe.DataFrame(result, ktdf.train, ktdf.encoders, ktdf.slice_id)

    def __repr__(self):
        if self.stl:
            return self.source
        else:
            return f'<Feature Constructor "{self.__name__}">'

    def __str__(self):
        return self.__name__


from . import stl


class FeatureSet:
    def __init__(self, fc_before, fc_after=stl.empty_like, df_input=None, target_column=None, name=None, description=None, desc=None, encoders=None):
        if type(fc_before) == list:
            self.fc_before = stl.concat(fc_before)
        elif type(fc_before) == tuple:
            self.fc_before = stl.compose(fc_before)
        else:
            self.fc_before = fc_before
        if type(fc_after) == list:
            self.fc_after = stl.concat(fc_after)
        elif type(fc_after) == tuple:
            self.fc_after = stl.compose(fc_after)
        else:
            self.fc_after = fc_after
        self.target_column = target_column
        self.encoders = (encoders if encoders is not None else dict())
        if df_input is not None:
            self.set_df(df_input)
        self._first_name = name
        self.__doc__ = None
        if desc is not None and description is not None:
            raise ValueError("desc is an alias of description. You can't use both")
        if description is not None:
            self.__doc__ = description
        elif desc is not None:
            self.__doc__ = desc

    def set_df(self, df_input):
        self.df_input = dataframe.DataFrame(df_input)
        self.df_input.train = True
        self.df_input.encoders = self.encoders
        self.df = self.fc_before(self.df_input)

    def __call__(self, df):
        ktdf = dataframe.DataFrame(df)
        ktdf.encoders = self.encoders
        return stl.merge([
            self.fc_before(ktdf),
            self.fc_after(ktdf)
        ])

    def __getitem__(self, idx):
        if isinstance(self.df_input, type(None)):
            raise AttributeError("Input DataFrame is not defined")
        return stl.merge([
            self.df.iloc[idx],
            self.fc_after(dataframe.DataFrame(self.df_input.iloc[idx], train=1))  # BUG: should have .train=True?
        ])  # made .train=1 only for preview purposes
        # actually, FS[a:b] functionality is made only for debug
        # why not write config.preview_call = 1 then?

    def empty_copy(self):
        return FeatureSet(self.fc_before,
                          self.fc_after,
                          target_column=self.target_column,
                          encoders=self.encoders,
                          name=self._first_name,
                          description=self.__doc__)

    def slice(self, idxs):
        return FeatureSlice(self, idxs)

    @property
    def target(self):
        if self.target_column:
            return self.df_input[self.target_column]
        else:
            raise AttributeError("Target column is not defined.")

    def __get_src(self, fc):
        if fc.__name__ == 'empty_like':
            return 'stl.empty_like'
        if not isinstance(fc, FeatureConstructor):
            return source_utils.get_source(fc)
        if fc.stl:
            return fc.source
        else:
            return fc.__name__

    @property
    def source(self):
        fc_before_source = self.__get_src(self.fc_before)
        fc_after_source = self.__get_src(self.fc_after)
        prefix = 'FeatureSet('
        fs_source = prefix + 'fc_before=' + fc_before_source + ',\n' \
                    + ' ' * len(prefix) + 'fc_after=' + fc_after_source + ',\n' \
                    + ' ' * len(prefix) +  'target_column=' + repr(self.target_column) + ')'
        return fs_source

    def recover_name(self):
        if self._first_name:
            return self._first_name
        ans = []
        frame = inspect.currentframe()
        tmp = {**frame.f_globals,                              # Please don't treat me as completely retarded
               **frame.f_back.f_globals,                       # At least because it works (almost always)
               **frame.f_back.f_back.f_globals,                # Actually, you don't even have to use this code
               **frame.f_back.f_back.f_back.f_globals,         # Just write FeatureSet(name="fs_1") and feel safe =|
               **frame.f_back.f_back.f_back.f_back.f_globals,
               **frame.f_back.f_back.f_back.f_back.f_back.f_globals,
               **frame.f_back.f_back.f_back.f_back.f_back.f_back.f_globals}

        for k, var in tmp.items():
            if isinstance(var, self.__class__):
                if hash(self) == hash(var) and k[0] != '_':
                    ans.append(k)
        if len(ans) != 1:
            print(f"The name cannot be uniquely defined. Possible options are: {ans}. Choosing {ans[0]}. You can set the name manually via FeatureSet(name=...) or using .set_name(...)")
        self._first_name = ans[0]
        return self._first_name

    @property
    def __name__(self):
        return self.recover_name()

    def set_name(self, name):
        self._first_name = name

    def describe(self):
        raise NotImplementedError


class FeatureSlice:
    def __init__(self, featureset, slice):
        self.featureset = featureset
        self.slice = slice
        self.slice_id = cache_utils.get_hash_slice(slice)
        self.first_level_encoders = self.featureset.encoders
        self.second_level_encoders = {}
        self.columns = None
        # self.df_input = copy(self.featureset.df_input)

    def __call__(self, df=None):
        if isinstance(df, type(None)):
            fsl_level_df = dataframe.DataFrame(self.featureset.df_input.iloc[self.slice],
                                               # ALERT: may face memory leak here
                                               slice_id=self.slice_id,
                                               train=True,
                                               encoders=self.second_level_encoders)
            result = stl.merge([
                self.featureset.df.iloc[self.slice],
                self.featureset.fc_after(fsl_level_df)
            ])
            self.columns = [i for i in result.columns if i != self.featureset.target_column]
            return result[self.columns]
        elif isinstance(df, slice) or isinstance(df, np.ndarray) or isinstance(df, list):
            fsl_level_df = dataframe.DataFrame(self.featureset.df_input.iloc[df],  # ALERT: may face memory leak here
                                               slice_id=self.slice_id,
                                               train=False,
                                               encoders=self.second_level_encoders)
            result = stl.merge([
                self.featureset.df.iloc[df],
                self.featureset.fc_after(fsl_level_df)
            ])
            for column in set(self.columns) - set(result.columns):
                result[column] = 0
            return result[self.columns]
        else:
            fs_level_df = dataframe.DataFrame(df,
                                              encoders=self.first_level_encoders)
            fsl_level_df = dataframe.DataFrame(df,
                                               encoders=self.second_level_encoders,
                                               slice_id=self.slice_id)
            result = stl.merge([
                self.featureset.fc_before(fs_level_df),  # uses FeatureSet-level encoders
                self.featureset.fc_after(fsl_level_df)  # uses FeatureSlice-level encoders
            ])
            for column in set(self.columns) - set(result.columns):
                result[column] = 0
            return result[self.columns]

    @property
    def target(self):
        return self.featureset.target.iloc[self.slice]

    def compress(self):
        self.featureset = self.featureset.empty_copy()


from collections import MutableSequence


class FeatureList(MutableSequence):
    def __init__(self):
        self.full_name = "kts.feature.storage.feature_list"  # such a hardcode
        self.names = [self.full_name]
        while self.names[-1].count('.'):
            self.names.append(self.names[-1][self.names[-1].find('.') + 1:])
        self.names.append('kts.features')
        while self.names[-1].count('.'):
            self.names.append(self.names[-1][self.names[-1].find('.') + 1:])
        self.functors = []
        self.name_to_idx = dict()

    def recalc(self):
        self.functors = []
        self.name_to_idx = dict()
        names = [obj for obj in caching.cache.cached_objs() if obj.endswith('_fc')]
        for idx, name in enumerate(names):
            functor = caching.cache.load_obj(name)
            self.functors.append(functor)
            self.name_to_idx[functor.__name__] = idx

    def __repr__(self):
        self.recalc()
        string = f"[{', '.join([f.__str__() for f in self.functors])}]"
        return string

    def __getitem__(self, key):
        self.recalc()
        if type(key) in [int, slice]:
            return self.functors[key]
        elif type(key) == str:
            return self.functors[self.name_to_idx[key]]
        else:
            raise TypeError('Index should be int, slice or str')

    def __delitem__(self, key):
        raise AttributeError('This object is read-only')

    def __setitem__(self, key, value):
        raise AttributeError('This object is read-only')

    def insert(self, key, value):
        raise AttributeError('This object is read-only')

    def define_in_scope(self, global_scope):
        self.recalc()
        for func in self.name_to_idx:
            for name in self.names:
                try:
                    exec(f"{func} = {name}['{func}']", global_scope)
                    break
                except BaseException:
                    pass

    def __len__(self):
        self.recalc()
        return len(self.functors)


feature_list = FeatureList()



import hashlib
import inspect
from itertools import zip_longest
import numpy as np
import time


def captcha():
    np.random.seed(int(time.time()))
    a, b = np.random.randint(5, 30, size=2)
    c = int(input(f"{a} + {b} = "))
    if a + b != c:
        return False
    return True


def list_hash(lst, length):
    return hashlib.sha256(repr(tuple(lst)).encode()).hexdigest()[:length]


def hash_str(a):
    return hashlib.sha256(a.encode()).hexdigest()


def extract_signature(func):
    args = inspect.getfullargspec(func).args
    defaults = inspect.getfullargspec(func).defaults
    values = {**inspect.currentframe().f_back.f_locals, **inspect.currentframe().f_back.f_back.f_locals}
    if defaults is None:
        defaults = []
    if args is None:
        args = []
    # print(inspect.getfullargspec(func))
    # print(values)
    sources = []
    for arg, default in list(zip_longest(args[::-1], defaults[::-1]))[::-1]:
        if values[arg] != default:
            if is_helper(values[arg]):
                arg_repr = values[arg].__name__
            else:
                arg_repr = repr(values[arg])
            sources.append(f'{arg}={arg_repr}')
    return ', '.join(sources)


def is_helper(func):
    return callable(func) \
           and '__name__' in dir(func) \
           and 'source' in dir(func) \
           and isinstance(func.source, str) \
           and 'def' in func.source  # genius




import numpy as np
import pandas as pd

from .experiment import experiment_list, Experiment
from .leaderboard import leaderboard
from ..pipeline import Pipeline
from ..modelling import Ensemble
from copy import deepcopy
from fastprogress import master_bar, progress_bar
from fastprogress.fastprogress import IN_NOTEBOOK

if IN_NOTEBOOK:
    from IPython.display import HTML


class Validator:
    def __init__(self, splitter, metric):
        self.splitter = splitter
        self.metric = metric
        
    def score(self, model, featureset, description=None, desc=None, **fit_params):
        try:
            _ = model.__name__
        except:
            raise AttributeError("Model must have .__name__ attribute to be validated")
        try:
            _ = model.source
        except:
            raise AttributeError("Model must have .source attribute to be validated")
        if desc is not None and description is not None:
            raise ValueError("desc is an alias of description. You can't use both")
        if desc is not None:
            description = desc
        pipelines = []
        scores = []
        y = featureset.target
        oofs = np.zeros_like(y, dtype=np.float)
        weights = np.zeros_like(y, dtype=np.float)
        model_name = f"{model.__name__}_x{self.splitter.get_n_splits()}-{featureset.__name__}"
        mb = master_bar(self.splitter.split(y, y),
                        total=self.splitter.get_n_splits(),
                        total_time=True)
        mb.write(f"Validation of {model_name}:")
        for idx_train, idx_test in mb:
            c_model = deepcopy(model)
            fsl = featureset.slice(idx_train)
            pl = Pipeline(c_model, fsl)
            pb = progress_bar(range(2), parent=mb)
            mb.child.comment = f'training...'
            pb.on_iter_begin()
            pb.update(0)
            fsl()
            try:
                pl.fit(eval_set=[(fsl(idx_test).values, featureset.target.values[idx_test])], **fit_params)
            except:
                mb.child.comment = 'failed to train with eval_set, training without it...'
                pl.fit(**fit_params)
            pb.update(1)
            mb.child.comment = 'validating...'
            pred = pl.predict(idx_test)
            pl.featureslice.compress()
            oofs[idx_test] = (weights[idx_test] * oofs[idx_test] + pred) / (weights[idx_test] + 1)
            weights[idx_test] += 1
            # print(featureset.target[idx_test].values[:10], pred[:10])
            score = self.metric(featureset.target.values[idx_test], pred)
            pipelines.append(pl)
            scores.append(score)
            pb.update(2)
            mb.first_bar.comment = f"{round(np.mean(scores), 7)}"
        final_ensemble = Ensemble(pipelines)
        final_ensemble = final_ensemble / len(pipelines)
        final_ensemble.__name__ = model_name
        score = np.mean(scores)
        std = np.std(scores)
        oofs = pd.DataFrame({'prediction': oofs})
        oofs.set_index(featureset.target.index, inplace=True)
        exp = Experiment(
                pipeline=final_ensemble,
                oofs=oofs,
                score=score,
                std=std,
                description=description,
                splitter=self.splitter,
                metric=self.metric)
        if IN_NOTEBOOK:
            mb.text = f"ID: {exp.identifier}<p>" + mb.text
            mb.out.update(HTML(mb.text))
        leaderboard.register(exp)
        return score




from . import cache_utils
from .. import config
import datetime
from glob import glob
import os
from .info import info
import pandas as pd
from .dataframe import DataFrame as KTDF

class Cache:
    """
    LRU cache for DataFrames and objects
    """

    def __init__(self):
        self.memory = dict()
        self.last_used = dict()
        self.edited_at = dict()
        self.current_volume = 0
        try:
            _ = info.memory_limit  # to check whether it exists
        except Exception:
            info.memory_limit = 4 * (1024 ** 3)  # 4 Gb

    @staticmethod
    def set_memory_limit(volume):
        """
        Sets a new memory limit in bytes
        :param volume: new memory limit
        :return:
        """
        info.memory_limit = volume

    def __release_volume(self, df):
        """
        Removes most unpopular dataframes until it is possible to cache given one
        :param df: dataframe
        :return:
        """

        items = sorted([(time, key) for (key, time) in self.last_used.items()])
        cur = 0
        while self.current_volume + cache_utils.get_df_volume(df) > info.memory_limit:
            key = items[cur][1]
            cur += 1
            self.current_volume -= cache_utils.get_df_volume(self.memory[key])
            self.memory.pop(key)
            self.last_used.pop(key)
            self.edited_at.pop(key)

    def is_cached_df(self, name):
        """
        Checks whether given df is cached
        :param name: name of dataframe
        :return: True or False (cache hit or miss)
        """
        # dict_name = name + '_df'
        # return dict_name in self.memory or os.path.exists(cache_utils.get_path_df(name))
        return os.path.exists(cache_utils.get_path_df(name))

    def cache_df(self, df, name):
        """
        Caches dataframe with given name
        :param df: df
        :param name: df name
        :return:
        """
        if self.is_cached_df(name):
            return
        if cache_utils.get_df_volume(df) > info.memory_limit:
            raise MemoryError

        dict_name = name + '_df'
        self.__release_volume(df)
        cache_utils.save_df(df, cache_utils.get_path_df(name))
        self.memory[dict_name] = df
        self.current_volume += cache_utils.get_df_volume(df)
        self.last_used[dict_name] = datetime.datetime.now()
        self.edited_at[dict_name] = cache_utils.get_time(cache_utils.get_path_df(name))

    def load_df(self, name):
        """
        Loads dataframe from cache
        :param name: name of df
        :return:
        """
        if not self.is_cached_df(name):
            raise KeyError("No such df in cache")

        dict_name = name + '_df'
        self.last_used[dict_name] = datetime.datetime.now()
        if dict_name in self.memory:
            if self.edited_at[dict_name] != cache_utils.get_time(cache_utils.get_path_df(name)):
                self.memory[dict_name] = cache_utils.load_df(cache_utils.get_path_df(name))
                self.edited_at[dict_name] = cache_utils.get_time(cache_utils.get_path_df(name))
            return self.memory[dict_name]
        else:
            tmp = cache_utils.load_df(cache_utils.get_path_df(name))
            self.__release_volume(tmp)
            self.memory[dict_name] = tmp
            self.edited_at[dict_name] = cache_utils.get_time(cache_utils.get_path_df(name))
            self.current_volume += cache_utils.get_df_volume(tmp)
            return tmp

    def remove_df(self, name):
        """
        Removes dataframe from cache
        :param name: name of df
        :return:
        """
        dict_name = name + '_df'
        if dict_name in self.memory:
            self.current_volume -= cache_utils.get_df_volume(self.memory[dict_name])
            self.memory.pop(dict_name)
        if dict_name in self.last_used:
            self.last_used.pop(dict_name)
        if dict_name in self.edited_at:
            self.edited_at.pop(dict_name)
        if os.path.exists(cache_utils.get_path_df(name)):
            os.remove(cache_utils.get_path_df(name))

    @staticmethod
    def cached_dfs():
        """
        Returns list of cached dataframes
        :return:
        """
        return [df.split('/')[-1][:-3] for df in
                sorted(glob(config.storage_path + '*' + '_df'), key=os.path.getmtime)]


    def is_cached_obj(self, name):
        """
        Checks whether object is in cache
        :param name: name of object
        :return: True or False (cache hit or miss)
        """
        # dict_name = name + '_obj'
        # return dict_name in self.memory or os.path.exists(cache_utils.get_path_obj(name))
        return os.path.exists(cache_utils.get_path_obj(name))

    def cache_obj(self, obj, name):
        """
        Caches object with given name
        :param obj: object
        :param name: object name
        :return:
        """
        if self.is_cached_obj(name):
            return

        dict_name = name + '_obj'
        self.memory[dict_name] = obj
        cache_utils.save_obj(obj, cache_utils.get_path_obj(name))
        self.edited_at[dict_name] = cache_utils.get_time(cache_utils.get_path_obj(name))

    def load_obj(self, name):
        """
        Loads object from cache
        :param name: name of object
        :return:
        """
        if not self.is_cached_obj(name):
            raise KeyError("No such object in cache")

        dict_name = name + '_obj'
        if dict_name in self.memory:
            if self.edited_at[dict_name] != cache_utils.get_time(cache_utils.get_path_obj(name)):
                self.memory[dict_name] = cache_utils.load_obj(cache_utils.get_path_obj(name))
                self.edited_at[dict_name] = cache_utils.get_time(cache_utils.get_path_obj(name))
            return self.memory[dict_name]
        else:
            tmp = cache_utils.load_obj(cache_utils.get_path_obj(name))
            self.memory[dict_name] = tmp
            self.edited_at[dict_name] = cache_utils.get_time(cache_utils.get_path_obj(name))
            return tmp

    def remove_obj(self, name):
        """
        Removes object from cache
        :param name: name of object
        :return:
        """
        dict_name = name + '_obj'
        if dict_name in self.memory:
            self.memory.pop(dict_name)
        if dict_name in self.last_used:
            self.last_used.pop(dict_name)
        if dict_name in self.edited_at:
            self.edited_at.pop(dict_name)
        if os.path.exists(cache_utils.get_path_obj(name)):
            os.remove(cache_utils.get_path_obj(name))

    @staticmethod
    def cached_objs():
        """
        Returns list of cached objects
        :return:
        """
        return [df.split('/')[-1][:-4] for df in
                sorted(glob(config.storage_path + '*' + '_obj'), key=os.path.getmtime)]


cache = Cache()


USER_SEP = '__USER__'


def save(obj, name):
    if name in ls():
        raise KeyError("You've already saved object with this name. If you want to overwrite it, first use kts.rm()")
    if isinstance(obj, KTDF) or isinstance(obj, pd.DataFrame):
        cache.cache_df(obj, name + USER_SEP)
    else:
        cache.cache_obj(obj, name + USER_SEP)


def ls():
    return [df.split('/')[-1][:-4 - len(USER_SEP)] for df in
            sorted(glob(config.storage_path + '*' + USER_SEP + '_obj'), key=os.path.getmtime)
           ] + [df.split('/')[-1][:-3 - len(USER_SEP)] for df in
            sorted(glob(config.storage_path + '*' + USER_SEP + '_df'), key=os.path.getmtime)]


def get_type(name):
    if cache.is_cached_obj(name + USER_SEP):
        return 'obj'
    else:
        return 'df'


def load(name):
    if name not in ls():
        raise KeyError("No such object in cache")
    if get_type(name) == 'df':
        return cache.load_df(name + USER_SEP)
    else:
        return cache.load_obj(name + USER_SEP)


def remove(name):
    if name not in ls():
        return
    if get_type(name) == 'df':
        cache.remove_df(name + USER_SEP)
    else:
        cache.remove_obj(name + USER_SEP)


rm = remove



from .. import config
import hashlib
import pandas as pd
import feather
import dill
from glob import glob
import os
import numpy as np
from ..utils import captcha


def clear_storage():
    from .caching import cache
    cache.memory.clear()
    if not captcha():
        print("You aren't smart enough to take such decisions")
        return
    for path in glob(config.storage_path + '*_*') + glob(config.source_path + '*'):
        print(f"deleting {path}")
        os.remove(path)


def get_hash_df(df):
    idx_hash = hashlib.sha256(pd.util.hash_pandas_object(df.index).values).hexdigest()

    sorted_cols = df.columns.sort_values()
    col_hash = hashlib.sha256(pd.util.hash_pandas_object(sorted_cols).values).hexdigest()
    try:
        hash_first, hash_last = pd.util.hash_pandas_object(df.iloc[[0, -1]][sorted_cols]).values
    except:
        hash_first = hashlib.sha256(df.iloc[[0]][sorted_cols].to_csv().encode('utf-8')).hexdigest()
        hash_last = hashlib.sha256(df.iloc[[-1]][sorted_cols].to_csv().encode('utf-8')).hexdigest()

    return hashlib.sha256(np.array([idx_hash, col_hash, hash_first, hash_last])).hexdigest()

    # return hashlib.sha256(pd.util.hash_pandas_object(df, index=True).values).hexdigest()


def get_hash_slice(idxs):
    if isinstance(idxs, slice):
        idxs = (-1337, idxs.start, idxs.stop, idxs.step)
    return hex(hash(frozenset(idxs)))[2:]


def get_df_volume(df):
    return df.memory_usage(index=True).sum()


def get_path_df(name):
    return config.storage_path + name + '_df'


def save_df(df, path):
    """
    Saves a dataframe as feather binary file. Adds to df and additional column filled
    with index values and having a special name.
    """
    # print('saving df', path)
    # try:
    #     print('enc', df.encoders)
    # except:
    #     print('enc: no attr')
    not_trivial_index = type(df.index) != pd.RangeIndex
    if not_trivial_index:
        index_name = f'{config.index_prefix}{df.index.name}'
        df[index_name] = df.index.values
        df.reset_index(drop=True, inplace=True)
    try:
        feather.write_dataframe(df, path)
    except Exception as e:
        print(e)
        try:
            df.to_parquet(path)
        except Exception as e1:
            print(e1)
            df.to_pickle(path)
    if not_trivial_index:
        df.set_index(index_name, inplace=True)
        df.index.name = df.index.name[len(config.index_prefix):]


def load_df(path):
    """
    Loads a dataframe from feather format and sets as index that additional
    column added with saving by save_df. Restores original name of index column.
    """
    try:
        tmp = feather.read_dataframe(path, use_threads=True)
    except:
        try:
            tmp = pd.read_parquet(path)
        except:
            tmp = pd.read_pickle(path)
    index_col = tmp.columns[tmp.columns.str.contains(config.index_prefix)]
    if any(index_col):
        tmp.set_index(index_col.values[0], inplace=True)
        tmp.index.name = tmp.index.name[len(config.index_prefix):]
    return tmp


def get_path_obj(name):
    return config.storage_path + name + '_obj'


def save_obj(obj, path):
    """
    Saves object
    """
    try:
        dill.dump(obj, open(path, 'wb'))
    except dill.PicklingError:
        if os.path.exists(path):
            os.remove(path)
            raise Warning(f'PicklingError occured, removing {path}')


def load_obj(path):
    """
    Loads object
    """
    return dill.load(open(path, 'rb'))


def get_path_info(name):
    return config.info_path + name + '_info'


def get_time(path):
    return int(os.path.getmtime(path))



from .storage import FeatureConstructor
import pandas as pd
import numpy as np
from ..storage.dataframe import DataFrame as KTDF
from ..zoo.cluster import KMeansFeaturizer
from sklearn.preprocessing import StandardScaler
from ..utils import list_hash, extract_signature, is_helper
from ..validation.leaderboard import leaderboard as lb


def wrap_stl_function(outer_function, inner_function):
    source = f'stl.{outer_function.__name__}({extract_signature(outer_function)})'
    fc = FeatureConstructor(inner_function, cache_default=False, stl=True)
    fc.source = source
    return fc


def empty_like(df):
    if isinstance(df, KTDF):
        return KTDF(df[[]].copy(), df.train, df.encoders, df.slice_id)
    else:
        return KTDF(df[[]].copy())


identity = FeatureConstructor(lambda df: df, cache_default=False)
identity.stl = True
identity.source = "stl.identity"


def merge(dfs):
    if len(dfs) == 1:
        return dfs[0]
    return pd.concat([df.df if isinstance(df, KTDF) else df for df in dfs], axis=1)


def column_selector(columns):
    def __col_selector(df):
        return df[columns]

    # return FeatureConstructor(__col_selector, cache_default=False)
    return wrap_stl_function(column_selector, __col_selector)


def column_dropper(columns):
    def __col_dropper(df):
        return df.drop(columns, axis=1)

    # return FeatureConstructor(__col_dropper, cache_default=False)
    return wrap_stl_function(column_dropper, __col_dropper)


def concat(funcs):
    def __concat(df):
        return merge([func(df) for func in funcs])

    fc = FeatureConstructor(__concat, cache_default=False)
    fc.source = f"stl.concat([{', '.join([func.__name__ if not func.stl else func.source for func in funcs])}])"
    fc.__name__ = f"concat_{''.join([func.__name__[:2] if not func.stl else func.__name__ for func in funcs])}"
    fc.stl = True
    return fc


def compose(funcs):
    def __compose(df):
        res = empty_like(df)
        for func in funcs:
            res = merge([res, func(res)])
        return res

    fc = FeatureConstructor(__compose, cache_default=False)
    fc.source = f"stl.compose([{', '.join([func.__name__ if not func.stl else func.source for func in funcs])}])"
    fc.__name__ = f"compose_{''.join([func.__name__[:2] if not func.stl else func.__name__ for func in funcs])}"
    fc.stl = True
    return fc


# TODO: reimplement using sklearn.preprocessing.OHE
def make_ohe(cols, sep='_ohe_'):
    def __make_ohe(df):
        return pd.get_dummies(df.df[cols], prefix_sep=sep, columns=cols)

    # return FeatureConstructor(__make_ohe, cache_default=False)
    return wrap_stl_function(make_ohe, __make_ohe)


# TODO: get helper-objects as args, not only functions, with fit() and transform() methods,
#       then call tfm.fit(df) and ..groupby().agg(tfm.transform)
#       will be useful for weighted target encoding,
#       like (df.sum() + C * global_mean) / (df.count() + C)
#       where global_mean is to be extracted inside of fit() call
# TODO: set global_answer var and .fillna(global_answer) for each pair of columns (for new classes)
def target_encoding(cols, target_cols, aggregation='mean', sep='_te_'):
    """ Template for creating target encoding FeatureConstructors

    :param cols: columns to encode
    :param target_cols: columns which values are aggregated
    :param aggregation: name of built-in pandas aggregation or @helper-function
    :param sep: string separator used for column naming
    :return: FeatureConstructor: df -> features
    """
    if type(cols) != list:
        cols = [cols]
    if type(target_cols) != list:
        target_cols = [target_cols]
    if isinstance(aggregation, str):
        aggregation_name = aggregation
    elif is_helper(aggregation):
        aggregation_name = aggregation.__name__
    else:
        raise TypeError('aggregation parameter should be either string or helper')

    def __target_encoding(df):
        res = empty_like(df)
        for target_col in target_cols:
            for col in cols:
                if df.train:
                    enc = df.groupby(col)[target_col].agg(aggregation)
                    global_answer = df[target_col].agg(aggregation)
                    df.encoders[f"__te_{col}_{target_col}_{aggregation_name}"] = enc
                    df.encoders[f"__te_{col}_{target_col}_{aggregation_name}_global_answer"] = global_answer
                else:
                    enc = df.encoders[f"__te_{col}_{target_col}_{aggregation_name}"]
                    global_answer = df.encoders[f"__te_{col}_{target_col}_{aggregation_name}_global_answer"]
                res[f"{col}{sep}{target_col}_{aggregation_name}"] = df[col].map(enc)
                res[f"{col}{sep}{target_col}_{aggregation_name}"].fillna(global_answer, inplace=True)
        return res

    # return FeatureConstructor(__target_encoding, cache_default=False)
    return wrap_stl_function(target_encoding, __target_encoding)


def target_encode_list(cols, target_col, aggregation='mean', prefix='me_list_'):
    def __target_encode_list(df):
        res = empty_like(df)
        filler = [np.nan]
        for col in cols:
            if df.train:
                enc = (pd.DataFrame([i if type(i) == list else [i] for i in df[col].tolist()], index=df[target_col])
                       .stack()
                       .reset_index(name=col)
                       [[col, target_col]]).groupby(col).agg(aggregation)
                df.encoders['__me_list_' + col] = enc
            else:
                enc = df.encoders['__me_list_' + col]
            res[prefix + col] = df[col].apply(lambda x: enc.loc[filter(lambda i: i in enc.index, x)].values.flatten() if type(x) == list else filler)
        return res

    # return FeatureConstructor(__target_encode_list, cache_default=False)
    return wrap_stl_function(target_encode_list, __target_encode_list)


import multiprocessing
import swifter


def apply(dataframe, function, **kwargs):
    """
    Applies function to dataframe faster.
    If n_threads is in kwargs and is greater than 1, applies by multiprocessing.
    :return: same as df.apply(function)
    """
    if 'n_threads' in kwargs:
        n_threads = kwargs.pop('n_threads')
    else:
        n_threads = 1
    if n_threads == 1:
        return dataframe.swifter.apply(function, **kwargs)

    def _apply_df(args):
        df, func, num, kw = args
        return num, df.swifter.apply(func, **kw)

    pool = multiprocessing.Pool(processes=n_threads)
    result = pool.map(_apply_df, [(d, function, i, kwargs) for i, d in enumerate(np.array_split(dataframe, n_threads))])
    pool.close()
    result = sorted(result, key=lambda x: x[0])
    return pd.concat([i[1] for i in result])


def get_categorical(df):
    if df.train:
        cat_features = [col for col in df.columns if df[col].dtype == object]
        df.encoders['__cat_features'] = cat_features
    else:
        cat_features = df.encoders['__cat_features']
    return cat_features


def get_numeric(df):
    if df.train:
        num_features = [col for col in df.columns if df[col].dtype != object]
        df.encoders['__num_features'] = num_features
    else:
        num_features = df.encoders['__num_features']
    return num_features


def discretize(cols, bins, prefix='disc_'):
    def __discretize(df):
        res = empty_like(df)
        for col in cols:
            if df.train:
                res[prefix + str(bins) + '_' + col], enc = pd.cut(df[col], bins, retbins=True)
                df.encoders[f'__disc_{bins}_{col}'] = enc
            else:
                enc = df.encoders[f'__disc_{bins}_{col}']
                res[prefix + str(bins) + '_' + col] = pd.cut(df[col], enc)
        return res

    # return FeatureConstructor(__discretize, cache_default=False)
    return wrap_stl_function(discretize, __discretize)


def discretize_quantile(cols, bins, prefix='disc_q_'):
    def __discretize_quantile(df):
        res = empty_like(df)
        for col in cols:
            if df.train:
                res[prefix + str(bins) + '_' + col], enc = pd.qcut(df[col], bins, retbins=True)
                df.encoders[f'__disc_q_{bins}_{col}'] = enc
            else:
                enc = df.encoders[f'__disc_q_{bins}_{col}']
                res[prefix + str(bins) + '_' + col] = pd.cut(df[col], enc)
        return res

    # return FeatureConstructor(__discretize_quantile, cache_default=False)
    return wrap_stl_function(discretize_quantile, __discretize_quantile)


def kmeans_encoding(cols, n_clusters, target_col=None, target_importance=5.0, prefix='km_', **kwargs):
    def __kmeans_encoding(df):
        res = empty_like(df)
        res_column_name = f'{prefix}{n_clusters}_{list_hash(cols, 5)}_{round(target_importance, 4)}'
        if df.train:
            encoder = KMeansFeaturizer(k=n_clusters, target_scale=target_importance, **kwargs)
            if target_col:
                res[res_column_name] = encoder.fit_transform(df[cols].values, df[target_col].values)
            else:
                res[res_column_name] = encoder.fit_transform(df[cols].values)
            df.encoders[f'__kmeans_feat_{n_clusters}_{list_hash(cols, 5)}_{round(target_importance, 4)}'] = encoder
        else:
            encoder = df.encoders[f'__kmeans_feat_{n_clusters}_{list_hash(cols, 5)}_{round(target_importance, 4)}']
            res[res_column_name] = encoder.transform(df[cols].values)
        return res

    # return FeatureConstructor(__kmeans_encoding, cache_default=False)
    return wrap_stl_function(kmeans_encoding, __kmeans_encoding)


def standardize(cols, prefix='std_'):
    def __standardize(df):
        res = empty_like(df)
        for col in cols:
            if df.train:
                encoder = StandardScaler()
                res[prefix + col] = encoder.fit_transform(df[[col]].values)
                df.encoders[f'__std_{col}'] = encoder
            else:
                encoder = df.encoders[f'__std_{col}']
                res[prefix + col] = encoder.transform(df[[col]].values)
        return res

    # return FeatureConstructor(__standardize, cache_default=False)
    return wrap_stl_function(standardize, __standardize)


def stack(ids):
    def __stack(df):
        oof_preds = merge([lb[id_exp].oofs for id_exp in ids])
        try:
            res = oof_preds[df.index]
            if res.isna().sum().sum() > 0:
                bad_indices = res.isna().sum(axis=1) > 0
                res[bad_indices] = merge([pd.DataFrame({id_exp: lb[id_exp].predict(df[bad_indices])}) for id_exp in ids])
        except:
            res = merge([pd.DataFrame({id_exp: lb[id_exp].predict(df)}) for id_exp in ids])
            res.set_index(df.index, inplace=True)
        return res

    return wrap_stl_function(stack, __stack)


# TODO: implement
def cv_apply(feature_constructor, split):
    raise NotImplementedError
    def __cv_apply(df):
        res = empty_like(df)



from ..modelling import *


class MultiClassifierMixin(Model):
    def predict(self, X, **kwargs):
        return self.predict_proba(X, **kwargs)

try:
    from xgboost import XGBClassifier as _XGBC

    class XGBClassifier(MultiClassifierMixin, _XGBC):
        short_name = 'xgb'
        tracked_params = [
            'base_score',
            'booster',
            'colsample_bylevel',
            'colsample_bytree',
            'gamma',
            'learning_rate',
            'max_delta_step',
            'max_depth',
            'min_child_weight',
            'missing',
            'n_estimators',
            'objective',
            'reg_alpha',
            'reg_lambda',
            'scale_pos_weight',
            'subsample'
        ]
except ImportError:
    pass


try:
    from lightgbm import LGBMClassifier as _LGBMC

    class LGBMClassifier(MultiClassifierMixin, _LGBMC):
        short_name = 'lgb'
        tracked_params = [
            'class_weight',
            'boost',
            'min_data_in_leaf',
            'bagging_freq',
            'subsample_freq',
            'bagging_fraction',
            'learning_rate',
            'reg_lambda',
            'feature_fraction',
            'min_sum_hessian_in_leaf',
            'boosting_type',
            'subsample_for_bin',
            'reg_alpha',
            'tree_learner',
            'colsample_bytree',
            'min_child_weight',
            'min_child_samples',
            'subsample',
            'max_depth',
            'boost_from_average',
            'num_leaves',
            'objective',
            'n_estimators',
            'min_split_gain',
        ]
except ImportError:
    pass


try:
    from catboost import CatBoostClassifier as _CBC

    class CatBoostClassifier(MultiClassifierMixin, _CBC):
        short_name = 'cb'
        tracked_params = [
            'iterations',
            'learning_rate',
            'l2_leaf_reg',
            'bootstrap_type',
            'bagging_temperature',
            'depth',
            'one_hot_max_size',
            'leaf_estimation_method',
            'nan_mode',
            'feature_border_type',
            'border_count',
            'max_depth',
            'colsample_bylevel',
            'eval_metric'
        ]
except ImportError:
    pass


from sklearn.ensemble import RandomForestClassifier as _RFC
class RandomForestClassifier(MultiClassifierMixin, _RFC):
    short_name = 'rf'
    tracked_params = [
        'bootstrap',
        'class_weight',
        'criterion',
        'max_depth',
        'max_features',
        'max_leaf_nodes',
        'min_impurity_decrease',
        'min_impurity_split',
        'min_samples_leaf',
        'min_samples_split',
        'min_weight_fraction_leaf',
        'n_estimators',
        'oob_score',
    ]


from sklearn.linear_model import LogisticRegression as _LR
class LogisticRegression(MultiClassifierMixin, _LR):
    short_name = 'lr'
    tracked_params = [
        'C',
        'penalty',
        'fit_intercept',
        'class_weight',
    ]



from ..storage import caching


# class Helper:
#     def __init__(self, function):
#         self.function = function
#         self.__name__ = function.__name__
#         self.source = source_utils.get_source(function)
#         self.__call__ = self.function.__call__
#
#     # # @functools.wraps(source_utils.get_source)
#     def __call__(self, *args, **kwargs):
#         return self.function(*args, **kwargs)
#
#     # def __getattribute__(self, item):
#     #     print(item)
#     #     return getattr(self.function, item)
#
#     def __repr__(self):
#         return self.function.__name__
#
#     def __str__(self):
#         return self.__name__
#
#

from collections import MutableSequence


class HelperList(MutableSequence):
    def __init__(self):
        self.full_name = "kts.feature.helper.helper_list"  # such a hardcode
        self.names = [self.full_name]
        while self.names[-1].count('.'):
            self.names.append(self.names[-1][self.names[-1].find('.') + 1:])
        self.names.append('kts.helpers')
        while self.names[-1].count('.'):
            self.names.append(self.names[-1][self.names[-1].find('.') + 1:])
        self.objects = []
        self.name_to_idx = dict()

    def recalc(self):
        self.objects = []
        self.name_to_idx = dict()
        names = [obj for obj in caching.cache.cached_objs() if obj.endswith('_helper')]
        for idx, name in enumerate(names):
            functor = caching.cache.load_obj(name)
            self.objects.append(functor)
            self.name_to_idx[functor.__name__] = idx

    def __repr__(self):
        self.recalc()
        string = f"[{', '.join([f.__name__ for f in self.objects])}]"
        return string

    def __getitem__(self, key):
        self.recalc()
        if type(key) in [int, slice]:
            return self.objects[key]
        elif type(key) == str:
            return self.objects[self.name_to_idx[key]]
        else:
            raise TypeError('Index should be int, slice or str')

    def __delitem__(self, key):
        raise AttributeError('This object is read-only')

    def __setitem__(self, key, value):
        raise AttributeError('This object is read-only')

    def insert(self, key, value):
        raise AttributeError('This object is read-only')

    def define_in_scope(self, global_scope):
        self.recalc()
        for func in self.name_to_idx:
            for name in self.names:
                try:
                    exec(f"{func} = {name}['{func}']", global_scope)
                    break
                except BaseException:
                    pass

    def __len__(self):
        self.recalc()
        return len(self.objects)


helper_list = HelperList()


import warnings
import re
from ..storage import cache


def is_identifier(string):
    return isinstance(string, str) and bool(re.match('[0-9A-F]{6}', string))


def is_list_of_identifiers(a):
    if not isinstance(a, list):
        return False
    for i in a:
        if not is_identifier(i):
            return False
    return True


def get_experiment(identifier):
    assert is_identifier(identifier), 'You can only get experiment by ID'
    names = [name for name in cache.cached_objs() if name.endswith('_exp') and name.startswith(identifier)]
    if len(names) == 0:
        raise KeyError('No such experiment found')
    elif len(names) == 1:
        return cache.load_obj(names[0])
    else:
        warnings.warn("I don't know what happened but there are several objects with this ID")
        return [cache.load_obj(name) for name in names]




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('VERSION', 'r') as f:
    VERSION = int(f.read().strip())

VERSION += 1

with open('VERSION', 'w') as f:
    f.write(str(VERSION))

setuptools.setup(
    name="kts",
    version=f"0.2.{VERSION}",
    author="Nikita Konodyuk",
    author_email="konodyuk@gmail.com",
    description="Competition-oriented framework for interactive feature engineering and building reproducible pipelines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/konodyuk/kts",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "mprop",
        "pandas",
        "numpy",
        "scikit-learn",
        "scikit-optimize",
        "matplotlib",
        "dill",
        "feather-format",
        "swifter",
        "kts-cli",
        "texttable",
        "fastprogress"
    ]
)



import pandas as pd
from ..storage import cache
from . import utils
from ..utils import captcha
from tqdm import tqdm

LB_DF_NAME = '__leaderboard'


class Leaderboard:
    def __init__(self):
        if LB_DF_NAME in cache.cached_dfs():
            self._df = cache.load_df(LB_DF_NAME)
        else:
            self._df = pd.DataFrame()
            cache.cache_df(self._df, LB_DF_NAME)

    def register(self, experiment):
        if experiment.__name__ + '_exp' in cache.cached_objs():
            return
        cache.cache_obj(experiment, experiment.__name__ + '_exp')
        self.add_row(experiment.as_df())

    def reload(self):
        self._df = cache.load_df(LB_DF_NAME)

    def add_row(self, row):
        self.reload()
        self._df = self._df.append(row)
        self._df = self._df.sort_values('Score', ascending=False)
        cache.remove_df(LB_DF_NAME)
        cache.cache_df(self._df, LB_DF_NAME)

    def __getattr__(self, item):
        return getattr(self._df, item)

    def __getitem__(self, key):
        if utils.is_identifier(key):
            return utils.get_experiment(key)
        elif utils.is_list_of_identifiers(key):
            return [utils.get_experiment(i) for i in key]
        else:
            res = self._df[key]
        if 'style' in dir(res):
            return res.style
        return res

    def _repr_html_(self):
        self.reload()
        return self._df.style._repr_html_()

    @property
    def df(self):
        return self._df.copy()

    def refresh(self):
        names = [name for name in cache.cached_objs() if name.endswith('_exp')]
        print(f'You want to refresh existing leaderboard.'
              f' It will require loading ALL ({len(names)}) existing experiments '
              f'and may take time. Do you want to continue?')
        if not captcha():
            return
        self._df = pd.DataFrame()
        cache.remove_df(LB_DF_NAME)
        cache.cache_df(self._df, LB_DF_NAME)
        for name in tqdm(names):
            self.add_row(cache.load_obj(name).as_df())
        print('Done')


leaderboard = Leaderboard()
